#include <include_closed/app/log/Logger.h>
#include <include_closed/app/App.h>
#include <include_closed/app/config/Config.h>
#include <include_closed/common/toolkit/vector/StrCpyVec.h>
#include <include_open/common/storage/Open_StorageErrors.h>
#include <include_open/common/toolkit/Open_StringTk.h>
#include <open/os/OsTypeConversion.h>
#include <open/filesystem/ProcFs.h>
#include "FhgfsOps.h"
#include "FhgfsOps_versions.h"

#include <linux/backing-dev.h>


/* some random number to identify fhgfs */
#define FHGFS_MAGIC  0x19830326


struct file_operations fhgfs_file_operations =
{
   .open       = FhgfsOps_open,
   .release    = FhgfsOps_release,
   .read       = FhgfsOps_read,
   .write      = FhgfsOps_write,
   .fsync      = FhgfsOps_fsync,
   .flush      = FhgfsOps_flush,
   .llseek     = FhgfsOps_llseek,
   .mmap       = FhgfsOps_mmap,

   #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
      .splice_read  = generic_file_splice_read,
      .splice_write = generic_file_splice_write,
   #endif // LINUX_VERSION_CODE

   #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,22)
      .sendfile   = generic_file_sendfile, // removed in 2.6.23 (replaced with splice)
   #endif // LINUX_VERSION_CODE
};

struct file_operations fhgfs_dir_operations =
{
   .open       = FhgfsOps_opendirIncremental,
   .release    = FhgfsOps_releasedir,
   .readdir    = FhgfsOps_readdirIncremental,
   .read       = generic_read_dir, // just returns the appropriate error code
};

struct inode_operations fhgfs_file_inode_operations =
{
   .getattr    = FhgfsOps_getattr,
   .setattr    = FhgfsOps_setattr,
   //.permission = FhgfsOps_permission, // default should be fine
   .truncate   = FhgfsOps_truncate,
};

struct inode_operations fhgfs_symlink_inode_operations =
{
   .readlink      = generic_readlink, //FhgfsOps_readlink, // default should be fine
   .follow_link   = FhgfsOps_follow_link,
   .put_link      = FhgfsOps_put_link,
};

static struct inode_operations fhgfs_dir_inode_operations =
{
   .lookup     = FhgfsOps_lookup,
   .create     = FhgfsOps_create,
   .link       = FhgfsOps_link,
   .unlink     = FhgfsOps_unlink,
   .symlink    = FhgfsOps_symlink,
   .mkdir      = FhgfsOps_mkdir,
   .rmdir      = FhgfsOps_rmdir,
   //.mknod      = FhgfsOps_mknod, // not supported currently
   .rename     = FhgfsOps_rename,
   .getattr    = FhgfsOps_getattr,
   .setattr    = FhgfsOps_setattr,
   //.permission = FhgfsOps_permission, // default should be fine
};

static struct dentry_operations fhgfs_dentry_operations =
{
   .d_revalidate   = FhgfsOps_revalidate,
   .d_delete       = FhgfsOps_deleteDentry,
};

static struct address_space_operations fhgfs_address_operations =
{
   .readpage      = FhgfsOps_readpage,
   .writepage     = FhgfsOps_writepage,
   .writepages    = generic_writepages,
   .prepare_write = FhgfsOps_prepare_write,
   .commit_write  = FhgfsOps_commit_write,
   .set_page_dirty = __set_page_dirty_nobuffers,
   //.bmap          = FhgfsOps_bmap, // required for swapon (if used directly - without loop mount)
   .direct_IO     = FhgfsOps_directIO,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
//static struct backing_dev_info fhgfs_backing_dev_info =
//{
//   .ra_pages   = 0, //VM_MAX_READAHEAD * 1024 / PAGE_CACHE_SIZE, /* read-ahead pages */
//   .state      = 0,
//   .capabilities  = BDI_CAP_MAP_COPY,
//   .unplug_io_fn  = default_unplug_io_fn,
//};
#endif // LINUX_VERSION_CODE

static struct super_operations fhgfs_sops =
{
   .statfs     = FhgfsOps_statfs,//simple_statfs,
   .drop_inode = generic_delete_inode,
   .put_super  = FhgfsOps_putSuper,
};


static struct file_system_type fhgfs_fs_type =
{
   .name       = "fhgfs",
   .owner      = THIS_MODULE,
   .get_sb     = __FhgfsOps_getSB,
   .kill_sb    = FhgfsOps_killSB,
   //.fs_flags   = FS_BINARY_MOUNTDATA, // not required currently
};


/**
 * Find out whether a directory entry exists. Add the dentry and create/associate the
 * corresponding inode in case it actually does exist.
 *
 * @param dentry the thing that the kernel is looking for
 */
struct dentry* FhgfsOps_lookup(struct inode* parentDir, struct dentry* dentry,
   struct nameidata* nameidata)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_lookup";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;
   FhgfsOpsErr statRes;
   fhgfs_stat fhgfsStat;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

   if(dentry->d_name.len > NAME_MAX)
   {
      retVal = -ENAMETOOLONG;
      goto clean_up;
   }

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   statRes = FhgfsOpsRemoting_stat(app, path, &fhgfsStat);

   if(statRes != FhgfsOpsErr_SUCCESS)
   { // error
      switch(statRes)
      {
         case FhgfsOpsErr_PATHNOTEXISTS:
         {
            dentry->d_op = &fhgfs_dentry_operations;
            d_add(dentry, NULL);
         } break;

         default:
         {
            retVal = FhgfsOpsErr_toSysErr(statRes);
         } break;
      }
   }
   else
   { // success (entry exists)
      struct kstat kstat;
      struct inode* newInode;

      OsTypeConv_kstatFhgfsToOs(&fhgfsStat, &kstat);

      kstat.ino = iunique(dentry->d_sb, 2);

      newInode = __FhgfsOps_newInode(parentDir->i_sb, &kstat, 0);
      if(!newInode)
         retVal = -EACCES;
      else
      {
         dentry->d_op = &fhgfs_dentry_operations;
         d_add(dentry, newInode);
      }
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

clean_up:
   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);


   return ERR_PTR(retVal);
}

/**
 * Note: Called when the dcache has a lookup hit (and wants to know whether the cache data
 * is still valid)
 */
int FhgfsOps_revalidate(struct dentry* dentry, struct nameidata* nameidata)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_revalidate";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int isValid = 0; // quasi-boolean
   struct inode* inode = dentry->d_inode;

   // note: isValid is quasi-boolean: 0 if entry invalidated, 1 if still valid.

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);


   if(inode)
   { // active dentry => remote-stat and local-verify

      __FhgfsOps_lockFhgfs(app); // B-K-LOCK

      if(is_bad_inode(inode) )
         LOG_DEBUG_FORMATTED(log, 4, logContext, "Revalidation request for bad inode: %s", path);
      else
      {
         isValid = (__FhgfsOps_refreshInodeUnlocked(dentry, path) == 0);
      }

      __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK
   }
   else
   { // negative dentry => we cannot update it => let's just say it's invalid
      isValid = 0;
   }


   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return isValid;
}

/**
 * Note: This is called from dput() when d_count is going to 0
 * Note2: We use this to unhash dentries with bad inodes
 */
int FhgfsOps_deleteDentry(struct dentry* dentry)
{
   // note: dput just wants to know from us whether or not it should delete the dentry
   int shouldBeDeleted = 0; // quasi-boolean
   struct inode* inode = dentry->d_inode;

   if(inode)
   {
      if(is_bad_inode(inode) )
         shouldBeDeleted = 1; // unhash the "bad inode"-dentry
   }
   else
   {
      // think about: we could unhash negative dentries here
   }

   return shouldBeDeleted;
}

loff_t FhgfsOps_llseek(struct file *file, loff_t offset, int origin)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_llseek";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   int64_t retVal = 0;
   struct inode *inode = file->f_mapping->host;


   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);


   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   switch(origin)
   {
      case SEEK_END:
      { // seek to position relative to end-of-file => get current file size first
         FhgfsOpsErr statRes;
         fhgfs_stat fhgfsStat;

         statRes = FhgfsOpsRemoting_stat(app, path, &fhgfsStat);
         if(statRes != FhgfsOpsErr_SUCCESS)
         { // error
            retVal = -EINVAL;
         }
         else
         {
            offset += fhgfsStat.size; //inode->i_size;
         }

      } break;

      case SEEK_CUR:
      { // seek to position relative to current offset
         offset += file->f_pos;
      } break;

      //case SEEK_SET: // seek to absolute position
   }


   if( (!retVal) && (offset >= 0) && (offset <= inode->i_sb->s_maxbytes) )
   { // new offset is valid
      if (offset != file->f_pos)
      {
         file->f_pos = offset;
         file->f_version = 0; // according to remote_llseek()
      }

      retVal = offset;
   }
   else
   { // new offset is invalid
      if(!retVal)
         retVal = -EINVAL;
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}


int FhgfsOps_getattr(struct vfsmount* mnt, struct dentry* dentry, struct kstat* kstat)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_getattr";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;


   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);


   retVal = __FhgfsOps_refreshInode(dentry, path);
   if(!retVal)
   {
      generic_fillattr(dentry->d_inode, kstat);

      if(!strcmp(path, "/") )
      { // grant access to root dir for everyone
         //kstat->mode = S_IFDIR | 0777;
         //stbuf->st_nlink = 2;
         //kstat->size = 161179;
         kstat->ino = 2; /* traditional root inode number */
      }

   }

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}


int FhgfsOps_setattr(struct dentry* dentry, struct iattr* iattr)
{
   // note: called for chmod(), chown() and utime()
   // note2: changed fields can be determined by the iattr->ia_valid flags

   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_setattr";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;
   fhgfs_bool refreshRequired = fhgfs_false;
   SettableFileAttribs fhgfsAttr;
   int validFhgfsAttribs;
   FhgfsOpsErr setAttrRes;
   struct inode* inode = dentry->d_inode;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);


   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   retVal = __FhgfsOps_refreshInodeUnlocked(dentry, path);
   if(retVal)
      goto refresh_and_unlock;

   if( (retVal = inode_change_ok(inode, iattr) ) < 0)
      goto refresh_and_unlock;

   OsTypeConv_iattrOsToFhgfs(iattr, &fhgfsAttr, &validFhgfsAttribs);

   setAttrRes = FhgfsOpsRemoting_setAttr(app, path, &fhgfsAttr, validFhgfsAttribs);

   if(setAttrRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(setAttrRes);
   }
   else
   { // success => handle the stuff that the server didn't handle

      refreshRequired = fhgfs_true;

      if(iattr->ia_valid & ATTR_SIZE)
      {
         filemap_write_and_wait(inode->i_mapping);

         inode->i_size = iattr->ia_size; // required for _truncate()
         /*retVal = */ FhgfsOps_truncate(inode);
         //if(retVal) // truncate does not have a return value
         //   goto refresh_and_unlock;

         // todo: uncomment the following line when we support mmap write in the future
         //retVal = vmtruncate(inode, iattr->ia_size);
      }
   }


refresh_and_unlock:

   if(refreshRequired)
      __FhgfsOps_refreshInodeUnlocked(dentry, path);

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return 0; // do not inform the application
}


/**
 * Note: Not really implemented (grants permission to everyone)
 */
int FhgfsOps_permission(struct inode* inode, int mask, struct nameidata* nameidata)
{
//   Logger* log = App_getLogger(Program_getApp() );
//   const char* logContext = "FhgfsOps_permission";
//
//   // note: nameidata might be null
//
//   struct dentry* dentry = d_find_alias(inode); // calls dget_locked()
//
//   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
//   char* pathStoreBuf;
//   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);
//
//   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);
//
//   // clean-up
//   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   // grant access for everyone if path exists
   return 0;

/*
    int res;

    res = access(path, mask);
    if (res == -1)
        return -errno;
*/
    // return 0;
}

int FhgfsOps_readlink(struct dentry* dentry, char __user* buf, int size)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_readlink";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

   // check user buffer
   if(unlikely(!access_ok(VERIFY_WRITE, buf, size) ) )
   {
      retVal = -EFAULT;
      goto clean_up;
   }

   retVal = FhgfsOpsHelper_readlink(app, path, buf, size);


   clean_up:
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

int FhgfsOps_opendirIncremental(struct inode* inode, struct file* file)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_opendirIncremental";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   int retVal;
   struct dentry* dentry = file->f_dentry;
   FsDirInfo* dirInfo;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   retVal = __FhgfsOps_refreshInodeUnlocked(dentry, path);
   if(!retVal)
   { // success (so this btw is still a directory)
      dirInfo = FsDirInfo_construct(app);
      FhgfsOps_setDirInfo(dirInfo, file);
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}


int FhgfsOps_readdirIncremental(struct file* file, void* buf, filldir_t filldir)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_readdirIncremental";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   int retVal = 0;
   FsDirInfo* dirInfo;
   struct dentry* dentry = file->f_dentry;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);


   dirInfo = FhgfsOps_getDirInfo(file);

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   //log.log(5, std::string("dirContents.size: ") + StringTk::intToStr(dirContents->size() ) );;
   //log.log(5, std::string("offset: ") + StringTk::intToStr(offset) );;

   if(file->f_pos < FsDirInfo_getOffset(dirInfo) )
   { // user seeked backwards => re-read the dir contents
      int rereadRes;
      loff_t currentOffset = (file->f_pos >= 2) ? (file->f_pos - 2) : 0; // avoid negative values;
         // -2 because of "." and ".."

      LOG_DEBUG_FORMATTED(log, 5, logContext, "Seeked backwards => re-reading directory: '%s' "
         "(old offset: %lld; new: %lld)", path, (long long)FsDirInfo_getOffset(dirInfo),
         (long long)file->f_pos);

      rereadRes = FhgfsOpsHelper_refreshDirInfoIncremental(app, path, currentOffset, dirInfo, fhgfs_true);

      //LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %d", (int)rereadRes);

      if(rereadRes)
      { // error occured
         __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

         // clean-up
         NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

         return rereadRes;
      }

      dirInfo = FhgfsOps_getDirInfo(file);
   }

   switch(file->f_pos)
   {
      case 0:
      {
         ino_t ino = dentry->d_inode->i_ino;
         if(filldir(buf, ".", 1, file->f_pos, ino, DT_DIR) < 0)
            break;
         file->f_pos++;

         LOG_DEBUG_FORMATTED(log, 5, logContext, "filled[%lld]: %s",
            (long long)0, ".");
      }
         /* fall through */
      case 1:
      {
         ino_t inoParent = parent_ino(dentry);
         if(filldir(buf, "..", 2, file->f_pos, inoParent, DT_DIR) < 0)
            break;
         file->f_pos++;

         LOG_DEBUG_FORMATTED(log, 5, logContext, "filled[%lld]: %s",
            (long long)1, "..");
      }
         /* fall through */
      default:
      {
         loff_t currentOffset = file->f_pos - 2; // -2 because of "." and ".."
         StrCpyVec* dirContents = FsDirInfo_getDirContents(dirInfo);
         //size_t dirContentsSize = StrCpyVec_length(dirContents);
         loff_t contentsOffset;
         size_t contentsLength;

         for( ; ; )
         {
            int refreshRes;
            char* currentName;

            refreshRes = FhgfsOpsHelper_refreshDirInfoIncremental(
               app, path, currentOffset, dirInfo, fhgfs_false);
            if(refreshRes)
            { // error occured
               retVal = refreshRes;
               break;
            }

            contentsOffset = FsDirInfo_getContentsOffset(dirInfo);
            contentsLength = StrCpyVec_length(dirContents);

            LOG_DEBUG_FORMATTED(log, 5, logContext,
               "stage: %d, currentOffset: %lld, contentsOffset: %lld, contentsLen: %lld",
               (int)FsDirInfo_getStage(dirInfo),
               (long long)currentOffset, (long long)contentsOffset, (long long)contentsLength);


            if( (FsDirInfo_getStage(dirInfo) == ListDirIncStage_END) &&
               (currentOffset >= (contentsOffset + contentsLength) ) )
            {
               LOG_DEBUG(log, 5, logContext, "reached end of dir");
               break;
            }

            currentName = StrCpyVec_at(
               dirContents, currentOffset - contentsOffset);

            //log.log(5, std::string("current iter value: ") + *iter);

            if(filldir(buf, currentName, strlen(currentName), file->f_pos, ~0, DT_UNKNOWN) < 0)
               break;

            LOG_DEBUG_FORMATTED(log, 5, logContext, "filled[%lld]: %s",
               (long long)currentOffset, currentName);

            file->f_pos++;
            currentOffset++;
         }
      }
   } // end of switch

   FsDirInfo_setOffset(dirInfo, file->f_pos);

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

/**
 * Note: This works for _opendir() and for _opendirIncremental().
 */
int FhgfsOps_releasedir(struct inode* inode, struct file* file)
{
   const char* logContext = "FhgfsOps_releasedir";

   //NoAllocBufferStore* bufStore = App_getPathBufStore(app); // debug in
   //char* pathStoreBuf; // debug in
   //char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf); // the old way

   FsDirInfo* dirInfo = FhgfsOps_getDirInfo(file);
   FsObjectInfo* fsObjectInfo = FhgfsOps_getObjectInfo(file);

   App* app = FsObjectInfo_getApp(fsObjectInfo);
   Logger* log = App_getLogger(app);

   //dput(dentry); // dput is probably wrong here (smbfs doesn't do it)

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", "n/a");


   dirInfo = FhgfsOps_getDirInfo(file);
   FsObjectInfo_virtualDestruct(fsObjectInfo);

   // clean-up
   //NoAllocBufferStore_addBuf(bufStore, pathStoreBuf); // debug in

   return 0;
}

// currently not supported and probably the code here is outdated!!
//int FhgfsOps_mknod(struct inode* dir, struct dentry* dentry, int mode, dev_t dev)
//{
//   Logger* log = App_getLogger(Program_getApp() );
//   const char* logContext = "FhgfsOps_mknod";
//
//   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
//   char* pathStoreBuf;
//   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);
//
//   int retVal = -ENOSPC;
//   struct inode* inode;
//
//   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);
//
//
//   inode = __FhgfsOps_newInode(dir->i_sb, mode, dev);
//
//   if(inode)
//   {
//      dentry->d_op = &fhgfs_dentry_operations;
//
//      d_instantiate(dentry, inode);
//      //dget(dentry);  /* Extra count - pin the dentry in core */ // dget could be unnecessary here
//
//      retVal = 0;
//
//      dir->i_ctime = dir->i_mtime = CURRENT_TIME;
//   }
//
//
//   // clean-up
//   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
//
//   return retVal; // do not allow this operation
//}

int FhgfsOps_mkdir(struct inode* dir, struct dentry* dentry, int mode)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_mkdir";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;
   FhgfsOpsErr mkRes;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

   mode |= S_IFDIR; // just make sure this is a dir

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   mkRes = FhgfsOpsRemoting_mkdir(app, path, current->fsuid, current->fsgid, mode);

   if(mkRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(mkRes);
   }
   else
   { // remote success => create the local inode
      retVal = __FhgfsOps_instantiateInode(dentry, path);
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

int FhgfsOps_rmdir(struct inode* dir, struct dentry* dentry)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_rmdir";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;
   FhgfsOpsErr rmRes;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   if(!d_unhashed(dentry) )
   {
      retVal = -EBUSY;
      goto unlock_and_exit;
   }

   rmRes = FhgfsOpsRemoting_rmdir(app, path);

   if(rmRes != FhgfsOpsErr_SUCCESS)
   {
      FhgfsOpsErr_toSysErr(rmRes);
   }
   else
   { // remote success
      // nothing to be done locally
   }


unlock_and_exit:

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

int FhgfsOps_create(struct inode* dir, struct dentry* dentry, int mode,
   struct nameidata* nameidata)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_create";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;
   FhgfsOpsErr mkRes;

   Logger_logFormatted(log, 4, logContext, "called. Path: %s", path);

   mode |=  S_IFREG; // just make sure we create a file here

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   mkRes = FhgfsOpsRemoting_mkfile(app, path, current->fsuid, current->fsgid, mode);

   if(mkRes != FhgfsOpsErr_SUCCESS)
   {
      if( (mkRes != FhgfsOpsErr_EXISTS) || (nameidata->intent.open.flags & O_EXCL) )
      { // Err_EXISTS is only an error if O_EXCL was specified

         retVal = FhgfsOpsErr_toSysErr(mkRes);
      }
   }

   if(!retVal) // keep this separate because of the O_EXCL special case above
   { // remote success => create local inode
      retVal = __FhgfsOps_instantiateInode(dentry, path);
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}


int FhgfsOps_unlink(struct inode* dir, struct dentry* dentry)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_unlink";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;
   FhgfsOpsErr unlinkRes;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   unlinkRes = FhgfsOpsRemoting_unlinkfile(app, path);

   if(unlinkRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(unlinkRes);
   }
   else
   { // remote success
      // nothing to be done locally
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}


int FhgfsOps_symlink(struct inode* dir, struct dentry* dentry, const char* to)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_symlink";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal;
   int mode = S_IFLNK | S_IRWXU | S_IRWXG | S_IRWXO;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s; To: %s", path, to);

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   retVal = FhgfsOpsHelper_symlink(app, path, to, current->fsuid, current->fsgid, mode);

   if(!retVal)
   { // remote success => create local inode
      retVal = __FhgfsOps_instantiateInode(dentry, path);
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

int FhgfsOps_rename(struct inode* inodeDirFrom, struct dentry* dentryFrom,
   struct inode* inodeDirTo, struct dentry* dentryTo)
{
   App* app = FhgfsOps_getApp(dentryFrom->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_rename";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBufFrom;
   char* pathStoreBufTo;

   char* pathFrom = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentryFrom, &pathStoreBufFrom);
   char* pathTo = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentryTo, &pathStoreBufTo);

   int retVal = 0;
   FhgfsOpsErr renameRes;

   Logger_logFormatted(log, 5, logContext, "called. From: %s; To: %s",
      pathFrom, pathTo);

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   if(dentryTo->d_inode)
      d_delete(dentryTo);

   renameRes = FhgfsOpsRemoting_rename(app, pathFrom, pathTo);

   if(renameRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(renameRes);
   }

   if(!retVal)
   { // remote success
      // nothing to be done here right now
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %d", (int)renameRes);

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBufFrom);
   NoAllocBufferStore_addBuf(bufStore, pathStoreBufTo);

   return retVal;
}

int FhgfsOps_link(struct dentry* dentryFrom, struct inode* inode, struct dentry* dentryTo)
{
   App* app = FhgfsOps_getApp(dentryFrom->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_link";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentryFrom, &pathStoreBuf);

   Logger_logFormatted(log, 2, logContext, "called - NOT IMPLEMENTED YET. From: %s", path);


   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return -EPERM; // do not allow creation of hardlink
}


int FhgfsOps_open(struct inode* inode, struct file* file)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Config* cfg = App_getConfig(app);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_open";

   //struct dentry* dentry = d_find_alias(inode); // calls __dget_locked()
   struct dentry* dentry = file->f_dentry;

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;
   int fhgfsOpenFlags;
   FsFileInfo* fileInfo;
   FhgfsOpsErr openRes;

   Logger_logFormatted(log, 4, logContext, "called. Path: %s", path);

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   fhgfsOpenFlags = OsTypeConv_openFlagsOsToFhgfs(file->f_flags);

   openRes = FhgfsOpsRemoting_openfile(app, path, fhgfsOpenFlags, &fileInfo);

   LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %d", (int)openRes);

   if(openRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(openRes);
   }
   else
   { // success => file is open (=> handle open flags)

      // handle O_TRUNC
      if(file->f_flags & O_TRUNC)
      {
         //inode->i_size = 0; // this is the way to tell truncate the desired trunc-size
         //FhgfsOps_truncate(inode);

         // old way (now done by server during open)
         //FhgfsOpsErr truncRes = FhgfsOpsRemoting_truncfile(app, path, 0);
         //if(truncRes != FhgfsOpsErr_SUCCESS)
         //   retVal = FhgfsOpsErr_toSysErr(truncRes);
      }

      // handle O_APPEND
      if(file->f_flags & O_APPEND)
         FsFileInfo_setAppending(fileInfo, fhgfs_true);

      // handle O_DIRECT + disabled caching
      if( (file->f_flags & O_DIRECT) || (!Config_getTuneFileCacheEnabled(cfg) ) )
         FsFileInfo_setAllowCaching(fileInfo, fhgfs_false);


      FhgfsOps_setFileInfo(fileInfo, file);
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

int FhgfsOps_release(struct inode* inode, struct file* file)
{
   const char* logContext = "FhgfsOps_release";

   //struct dentry* dentry = d_find_alias(inode); // not working here

   //NoAllocBufferStore* bufStore = App_getPathBufStore(app); // debug in
   //char* pathStoreBuf; // debug in
   //char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf); // the old way

   int retVal = 0;
   FhgfsOpsErr closeRes;
   FsFileInfo* fileInfo = FhgfsOps_getFileInfo(file);
   FsObjectInfo* fsObjectInfo = FhgfsOps_getObjectInfo(file);
   FhgfsOpsErr flushRes;

   App* app = FsObjectInfo_getApp(fsObjectInfo);
   Logger* log = App_getLogger(app);

   Logger_logFormatted(log, 4, logContext, "called. Path: %s", "n/a");

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   if(unlikely(!fileInfo) )
   { // invalid file
      retVal = -EBADF;
      goto unlock_and_cleanup;
   }

   filemap_write_and_wait(inode->i_mapping);

   flushRes = FhgfsOpsHelper_flushCache(app, fileInfo, fhgfs_true);
   if(flushRes != FhgfsOpsErr_SUCCESS)
   { // error (error code might be overridden by closeRes)
      retVal = FhgfsOpsErr_toSysErr(flushRes);
   }

   closeRes = FhgfsOpsRemoting_closefile(fileInfo);

   LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %d", (int)closeRes);

   if(closeRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(closeRes);
   }

   // note: we free the fileInfo no matter whether the communication succeeded or not
   //    (because _release() won't be called again even if it didn't succeed)

   FsObjectInfo_virtualDestruct( (FsObjectInfo*)fileInfo);
   FhgfsOps_setFileInfo( (FsFileInfo*)NULL, file);


unlock_and_cleanup:
   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   // clean-up
   //NoAllocBufferStore_addBuf(bufStore, pathStoreBuf); // debug in

   return retVal;
}

ssize_t FhgfsOps_read(struct file* file, char __user *buf, size_t size, loff_t* offsetPointer)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);

   ssize_t readRes;

   #ifdef LOG_DEBUG_MESSAGES

      Logger* log = App_getLogger(app);
      const char* logContext = "FhgfsOps_read";

      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s", path);

      LOG_DEBUG_FORMATTED(log, 5, logContext, "size/offset: %lld/%lld",
         (long long)size, (long long)*offsetPointer);

      // clean-up
      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   #endif // LOG_DEBUG_MESSAGES


   // check user buffer
   if(unlikely(!access_ok(VERIFY_WRITE, buf, size) ) )
      return -EFAULT;

   readRes = FhgfsOpsHelper_readCached(
      app, buf, size, *offsetPointer, FhgfsOps_getFileInfo(file) );
   //readRes = FhgfsOpsRemoting_readfile(
   //   buf, size, *offsetPointer, FhgfsOps_getFileInfo(file) );
   //readRes = FhgfsOpsRemoting_readfileThreaded(
   //   buf, size, *offsetPointer, FhgfsOps_getFileInfo(file) );

   if(unlikely(readRes < 0) )
   { // error
      return FhgfsOpsErr_toSysErr(-readRes);
   }

   *offsetPointer += readRes;

   return readRes;
}

ssize_t FhgfsOps_write(struct file* file, const char __user *buf, size_t size,
   loff_t* offsetPointer)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);

   ssize_t writeRes;
   FsFileInfo* fileInfo;

   #ifdef LOG_DEBUG_MESSAGES

      Logger* log = App_getLogger(app);
      const char* logContext = "FhgfsOps_write";

      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s (size: %lld; offset: %lld)",
         path, (long long)size, (long long)*offsetPointer);

      // clean-up
      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   #endif // LOG_DEBUG_MESSAGES


   // check user buffer
   if(unlikely(!access_ok(VERIFY_READ, buf, size) ) )
      return -EFAULT;

   fileInfo = FhgfsOps_getFileInfo(file);

   if(FsFileInfo_getAppending(fileInfo) )
   { // appending => move file offset to end-of-file before writing
      FhgfsOpsErr statRes;
      fhgfs_stat fhgfsStat;

      NoAllocBufferStore* bufStore = App_getPathBufStore(app);
      char* pathStoreBuf;
      char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

      statRes = FhgfsOpsRemoting_stat(app, path, &fhgfsStat);
      if(statRes != FhgfsOpsErr_SUCCESS)
         return FhgfsOpsErr_toSysErr(statRes);

      *offsetPointer = fhgfsStat.size;

      // clean-up
      NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   }

   writeRes = FhgfsOpsHelper_writeCached(app, buf, size, *offsetPointer, fileInfo);
   //writeRes = FhgfsOpsRemoting_writefile(buf, size, *offsetPointer, fileInfo);
   //writeRes = FhgfsOpsRemoting_writefileThreaded(buf, size, *offsetPointer, fileInfo);

   if(unlikely(writeRes < 0) )
   { // error
      return FhgfsOpsErr_toSysErr(-writeRes);
   }

   *offsetPointer += writeRes;

   return writeRes;
}

void FhgfsOps_truncate(struct inode* inode)
{
   // note: i_size of the inode is set to the desired size

   struct dentry* dentry = d_find_alias(inode); // calls dget_locked

   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_truncate";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;
   FhgfsOpsErr truncRes;

   loff_t size = inode->i_size;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);


   truncRes = FhgfsOpsRemoting_truncfile(app, path, size);

   if(truncRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(truncRes);
   }

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   dput(dentry);

   return; // retVal; // how do we return a result?!
}


int FhgfsOps_fsync(struct file* file, struct dentry* dentry, int datasync)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_fsync";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int retVal = 0;
   struct inode* inode = file->f_dentry->d_inode;
   FhgfsOpsErr flushRes;
   FhgfsOpsErr fsyncRes;
   FsObjectInfo* fsObjectInfo;
   FsFileInfo* fileInfo;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

   fsObjectInfo = FhgfsOps_getObjectInfo(file);

   // syncing somehing other than a file?
   if(FsObjectInfo_getObjectType(fsObjectInfo) != FsObjectType_FILE)
      goto clean_up;

   fileInfo = FhgfsOps_getFileInfo(file);

   //printk_fhgfs_debug(FHGFS_KERN_INFO, "waiting for fsync page writes\n"); // debug in (delete me)
   retVal = filemap_write_and_wait(inode->i_mapping);
   //printk_fhgfs_debug(FHGFS_KERN_INFO, "fsync page writes complete\n"); // debug in (delete me)
   if(retVal)
      goto clean_up;


   flushRes = FhgfsOpsHelper_flushCache(app, fileInfo, fhgfs_false);
   if(flushRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(flushRes);
      goto clean_up;
   }

   fsyncRes = FhgfsOpsRemoting_fsyncfile(fileInfo);
   retVal = FhgfsOpsErr_toSysErr(fsyncRes);

   // clean-up
clean_up:
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

int FhgfsOps_mmap(struct file* file, struct vm_area_struct* vma)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_mmap";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   int retVal;
   FhgfsOpsErr statRes;
   fhgfs_stat fhgfsStat;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);


   LOG_DEBUG_FORMATTED(log, 5, logContext, "updating inode... Path: %s", path);


   statRes = FhgfsOpsRemoting_stat(app, path, &fhgfsStat);
   if(statRes != FhgfsOpsErr_SUCCESS)
   {
      retVal = FhgfsOpsErr_toSysErr(statRes);
   }
   else
   { // success
      struct kstat kstat;
      OsTypeConv_kstatFhgfsToOs(&fhgfsStat, &kstat);

      __FhgfsOps_applyStatDataToInode(&kstat, file->f_dentry->d_inode);

      LOG_DEBUG_FORMATTED(log, 5, logContext, "inode updated. Path: %s", path);


      //retVal = generic_file_readonly_mmap(file, vma);
      retVal = generic_file_mmap(file, vma);
   }

   LOG_DEBUG_FORMATTED(log, 5, logContext, "result: %d", retVal);


   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

/*
 * Note: Called with the page locked and we need to unlock it when done.
 *
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOps_readpage(struct file* file, struct page* page)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_readpage";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   int retVal;

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s", path);
   IGNORE_UNUSED_VARIABLE(logContext);
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(path);


   retVal = __FhgfsOps_readpage_sync(file, page);

   unlock_page(page);

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

/**
 * @return 0 on success
 */
int FhgfsOps_writepage(struct page *page, struct writeback_control *wbc)
{
   // Note: this is a writeback method, so we have no file handle here! (only the inode)

   // note: make sure that write-mapping is enabled in fhgfsops_mmap!!!


   App* app;
   Logger* log;
   const char* logContext = "FhgfsOps_writepage";

   //NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   //char* pathStoreBuf;
   //char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   int retVal = 0;
   struct address_space* mapping = page->mapping;
   struct inode* inode;

   if(!mapping || !mapping->host)
      return -EFAULT;

   inode = mapping->host;
   app = FhgfsOps_getApp(inode->i_sb);
   log = App_getLogger(app);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s", "n/a"/*path*/);
   IGNORE_UNUSED_VARIABLE(logContext);


   page_cache_get(page);
   if(!PageUptodate(page) )
      LOG_DEBUG_FORMATTED(log, 5, logContext, "page not uptodate");

   // set writeback-flag, and clear dirty-flag.
   // note: the alternative is redirty_page_for_writepage() in case of an error

   set_page_writeback(page);

   retVal = __FhgfsOps_writepage_async(page, 0, PAGE_CACHE_SIZE);

   SetPageUptodate(page);
   unlock_page(page);
   end_page_writeback(page);
   page_cache_release(page);


   // clean-up
   //NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return retVal;
}

/**
 * Note: We never return an error code here, currently.
 *
 * @param from page inner offset start
 * @param to page inner offset end
 * @return 0 on success
 */
int FhgfsOps_prepare_write(struct file* file, struct page* page, unsigned from, unsigned to)
{
   // Note: Make sure that write-mapping is enabled in fhgfsOps_mmap!!!

   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_prepare_write";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   int retVal = 0;

   loff_t fileSize;
   loff_t offset;


   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s; from: %u; to: %u", path, from, to);
   IGNORE_UNUSED_VARIABLE(logContext);
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(path);

   if(PageUptodate(page) )
      goto clean_up;

   if( (to == PAGE_CACHE_SIZE) && (from == 0) )
   { // we're writing a full page
      // full page => will be up to date => no need to read-update the page from the server
      SetPageUptodate(page);
      goto clean_up;
   }

   // we're not writing a full page => try to read-update the page

   offset = (loff_t)page->index << PAGE_CACHE_SHIFT;
   fileSize = i_size_read(page->mapping->host);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "fileSize: %lld; offset: %lld",
      (long long)fileSize, (long long)offset);

   if( (offset >= fileSize) ||
       ( (from == 0) && (offset + to) >= fileSize) )
   {
      // we don't need to read data beyond the end of the file
      //    => zero it, and set the page uptodate
      simple_prepare_write(file, page, from, to);
      SetPageUptodate(page);
   }
   else
   if( (file->f_flags & O_ACCMODE) != O_WRONLY)
   {
      // file open for reading also => read-update the page */
      retVal = __FhgfsOps_readpage_sync(file, page);
   }
   else
   { // Write-only file handle => we cannot use it to read-update this page.
      // Trying to use a different currently open file handle would bring race conditions.
      // In either case this will be written out by commit_write, so it's fine.
   }


   // clean-up
clean_up:
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   return 0;
}

/**
 * @param from page inner offset start
 * @param to page inner offset end
 *
 * @return 0 in success
 */
int FhgfsOps_commit_write(struct file* file, struct page* page, unsigned from, unsigned to)
{
   // note: make sure that write-mapping is enabled in fhgfsops_mmap!!!

   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_commit_write";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, file->f_dentry, &pathStoreBuf);

   int retVal = 0;
   struct inode* inode = page->mapping->host;
   loff_t pageStart = (loff_t)page->index << PAGE_CACHE_SHIFT;
   loff_t endPos = pageStart + to;

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s", path);
   IGNORE_UNUSED_VARIABLE(path);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "pageStart: %lld; from: %u; to: %u",
      (long long)pageStart, from, to);


   // check fileSize and update it to page endPos if required
   spin_lock(&inode->i_lock);
   if(endPos > inode->i_size)
   {
      i_size_write(inode, endPos);
   }
   spin_unlock(&inode->i_lock);

   // write the page directly or mark it for async writing
   if(!PageUptodate(page) )
   { // write now
      if(unlikely(to < from) )
      {
         Logger_logFormatted(log, 2, logContext, "Illegal offsets. from: %d; to: %d", from, to);
         goto clean_up;
      }

      retVal = __FhgfsOps_writepage_sync(file, page, from, to-from);
      if(retVal > 0)
         retVal = 0;
   }
   else
   { // mark for async writing via writepage
      set_page_dirty(page);
   }


   // clean-up
clean_up:
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "complete. retVal: %d", retVal);

   return retVal;
}


/**
 * Called by the VFS to map a logical block offset within object to
 * physical block number.
 *
 * @param mapping address space info
 * @param block  the block to be mapped
 * @return disk address for the block or 0 on hole or error
 */
sector_t FhgfsOps_bmap(struct address_space* mapping, sector_t block)
{
   //printk_fhgfs(FHGFS_KERN_INFO, "FhgfsOps_bmap called. block: %lld\n", (long long)block); // debug in

   // note: +1 to avoid returning 0 for the very first block
   return block+1;
}

/**
 * This method must be present because otherwise the kernel rejects open with O_DIRECT in
 * fs/open.c. However, it should actually never be called.
 *
 * @param rw whether this is a read or a write
 * @param iocb I/O control block
 * @param iov I/O buffer vectors (array)
 * @param pos file offset
 * @param nr_segs length of iov array
 *
 */
ssize_t FhgfsOps_directIO(int rw, struct kiocb *iocb, const struct iovec *iov, loff_t pos,
   unsigned long nr_segs)
{
   printk_fhgfs_debug(FHGFS_KERN_INFO, "FhgfsOps_directIO was called (should never happen). "
      "File: %s\n", iocb->ki_filp->f_path.dentry->d_name.name);

   return -EINVAL;
}


/**
 * Creates a new inode, inits it from the kstat, inits the ops (depending on the mode)
 * and hashes it.
 * Note: Make sure everything is set in the kstat _before_ you call this, because we hash
 * the inode in here.
 * Note2: This should not be called directly. Instead, do the remote stuff first and then
 * use the _instantiateInode()-wrapper
 *
 * @param kstat must have a valid .ino (inode number)
 * @param dev set to 0 if not required
 * @return NULL if not successful
 */
struct inode* __FhgfsOps_newInode(struct super_block* sb, struct kstat* kstat, dev_t dev)
{
   struct inode* inode = new_inode(sb);
   if(!inode)
      return inode;

   __FhgfsOps_applyStatDataToInode(kstat, inode);

   inode->i_ino = kstat->ino;


   switch (kstat->mode & S_IFMT)
   {
      case S_IFREG:
      {
         inode->i_op = &fhgfs_file_inode_operations;
         inode->i_fop = &fhgfs_file_operations;
         inode->i_data.a_ops = &fhgfs_address_operations;
         //inode->i_mapping = &inode->i_data; // do we need this?

         #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
         //inode->i_mapping->backing_dev_info = &fhgfs_backing_dev_info; // debug in
         #endif // LINUX_VERSION_CODE
      } break;
      case S_IFDIR:
      {
         inode->i_op = &fhgfs_dir_inode_operations;
         inode->i_fop = &fhgfs_dir_operations;

         /* directory inodes start off with i_nlink == 2 (for "." entry) */
         //inode->i_nlink = 2; // not necessary - nodes send the nlink count
      } break;
      case S_IFLNK:
      {
         //inode->i_op = &page_symlink_inode_operations;
         inode->i_op = &fhgfs_symlink_inode_operations;
      } break;

      default:
      {
         init_special_inode(inode, kstat->mode, dev);
      } break;
   }

   insert_inode_hash(inode);

   return inode;
}

/**
 * Retrieves the attribs from the storage node, generates an inode number (ID) and instantiates
 * a local version of the inode for the dentry.
 *
 * Note: This is a wrapper for _newInode()
 *
 * @param resolved path for the dentry
 * @return 0 on success, negative linux error code otherwise
 */
int __FhgfsOps_instantiateInode(struct dentry* dentry, char* path)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   int retVal = 0;
   fhgfs_stat fhgfsStat;
   FhgfsOpsErr statRes;

   statRes = FhgfsOpsRemoting_stat(app, path, &fhgfsStat);

   if(statRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(statRes);
   }
   else
   { // success (entry exists on server)
      struct kstat kstat;
      struct inode* newInode;

      OsTypeConv_kstatFhgfsToOs(&fhgfsStat, &kstat);

      kstat.ino = iunique(dentry->d_sb, 2);

      newInode = __FhgfsOps_newInode(dentry->d_sb, &kstat, 0);
      if(!newInode)
         retVal = -EACCES;
      else
      { // new inode created
         d_instantiate(dentry, newInode);
      }
   }

   return retVal;
}


/**
 * Note: Reads a locked page and unlocks it when done
 *
 * @return 0 on success, negative linux error code otherwise
 */
int __FhgfsOps_readpage_sync(struct file* file, struct page *page)
{
   const char* logContext = "__FhgfsOps_readpage_sync";

   int retVal = 0;
   char* buf;
   loff_t offset = (loff_t)page->index << PAGE_CACHE_SHIFT;
   ssize_t readRes;

   FsObjectInfo* fsObjectInfo = FhgfsOps_getObjectInfo(file);
   FsFileInfo* fileInfo = FhgfsOps_getFileInfo(file);
   App* app = FsObjectInfo_getApp(fsObjectInfo);
   Logger* log = App_getLogger(app);


   page_cache_get(page);
   buf = kmap(page);


   LOG_DEBUG_FORMATTED(log, 5, logContext, "offset: %lld", (long long)offset);
   IGNORE_UNUSED_VARIABLE(logContext);
   IGNORE_UNUSED_VARIABLE(log);

   readRes = FhgfsOpsRemoting_readfile(buf, PAGE_CACHE_SIZE, offset, fileInfo);
   if(readRes < 0)
   { // error occured
      retVal = -EREMOTEIO;

      goto clean_up;
   }

   if(readRes < PAGE_CACHE_SIZE)
   { // did not receive the whole page => zero the remaining bytes
      memset(&buf[readRes], 0, PAGE_CACHE_SIZE - readRes);
   }

   flush_dcache_page(page);
   SetPageUptodate(page);

   retVal = 0;

clean_up:
   kunmap(page);
   page_cache_release(page);

   return retVal;
}

/**
 * @param pageoffset is the inner-page offset (relative to the page start)
 */
int __FhgfsOps_writepage_sync(struct file* file, struct page* page, unsigned pageOffset,
   unsigned count)
{
   const char* logContext = "__FhgfsOps_writepage_sync";

   int retVal = 0;

   char* buf = kmap(page);
   loff_t offset = ( (loff_t)page->index << PAGE_CACHE_SHIFT) + pageOffset;
   ssize_t writeRes;

   FsObjectInfo* fsObjectInfo = FhgfsOps_getObjectInfo(file);
   FsFileInfo* fileInfo = FhgfsOps_getFileInfo(file);
   App* app = FsObjectInfo_getApp(fsObjectInfo);
   Logger* log = App_getLogger(app);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "offset: %lld", (long long)offset);
   IGNORE_UNUSED_VARIABLE(logContext);
   IGNORE_UNUSED_VARIABLE(log);

   writeRes = FhgfsOpsRemoting_writefile(&buf[pageOffset], count, offset, fileInfo);
   if(writeRes < 0)
   { // error occured
      retVal = -EREMOTEIO;
   }

   // clean-up
   kunmap(page);

   return retVal;
}

/**
 * @return 0 on success
 */
int __FhgfsOps_writepage_async(struct page* page, unsigned from, unsigned to)
{
   App* app;
   Logger* log;
   const char* logContext = "__FhgfsOps_writepage_async";

   NoAllocBufferStore* bufStore;
   char* pathStoreBuf = NULL;
   char* path;

   struct address_space* mapping = page->mapping;
   struct inode* inode;
   struct dentry* dentry = NULL;
   loff_t offset = (loff_t)page->index << PAGE_CACHE_SHIFT;
   char* buf = NULL;
   int retVal = 0;
   ssize_t writeRes;

   if(!mapping || !mapping->host)
   {
      retVal = -EFAULT;
      goto err_exit;
   }

   inode = mapping->host;
   dentry = d_find_alias(inode); // calls dget_locked
   app = FhgfsOps_getApp(inode->i_sb);
   log = App_getLogger(app);
   bufStore = App_getPathBufStore(app);

   if(!dentry)
   {
      printk_fhgfs(FHGFS_KERN_INFO, "Unable to find dentry for inode during async page flush\n");
      retVal = -EFAULT;
      goto err_exit;
   }

   path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "called. Path: %s; offset: %lld",
      path, (long long)offset);
   IGNORE_UNUSED_VARIABLE(logContext);
   IGNORE_UNUSED_VARIABLE(log);

   offset += (loff_t)from;
   buf = kmap(page);

   if( (to > PAGE_CACHE_SIZE) || (from > to) )
   {
      kunmap(page);
      retVal = -EIO;
      goto clean_up;
   }

   // racing with truncate?
   if(offset > mapping->host->i_size)
   {
      kunmap(page);
      retVal = 0; // don't care
      goto clean_up;
   }

   // check to make sure that we are not extending the file
   if(mapping->host->i_size - offset < (loff_t)to)
      to = (unsigned)(mapping->host->i_size - offset);

   writeRes = FhgfsOpsHelper_writeStateless(app, path, &buf[from], to-from, offset);
   if(writeRes == -1)
      retVal = -EREMOTEIO;

   kunmap(page);

   // clean-up
clean_up:
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);
   dput(dentry);

err_exit:

   return retVal;
}


/**
 * Called when we want to check whether the inode has changed on the server (or to update
 * the inode attribs after we've made changes to a file/dir).
 * If it has changed, we must invalidate our local caches.
 *
 * Note: This version does not acquire the B-K-Lock.
 *
 * @param dentry the dentry for the inode that should be revalidated
 * @param path the resolved path (relative to mount point) for the dentry
 * @return 0 on success (validity), negative linux error code if no longer valid
 */
int __FhgfsOps_refreshInodeUnlocked(struct dentry* dentry, char* path)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   int retVal = 0;
   FhgfsOpsErr statRes;
   fhgfs_stat fhgfsStat;
   struct inode* inode = dentry->d_inode;


   statRes = FhgfsOpsRemoting_stat(app, path, &fhgfsStat);

   if(statRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(statRes);
   }
   else
   { // success (entry exists)
      struct kstat kstat;
      OsTypeConv_kstatFhgfsToOs(&fhgfsStat, &kstat);

      if(inode->i_ino == 2)
      { // root node
         __FhgfsOps_applyStatDataToInode(&kstat, inode);
      }
      else
      { // any existing node => check whether it is still the same object type
         if( (inode->i_mode & S_IFMT) == (kstat.mode & S_IFMT) )
         { // mode(-type) is still the same => update attribs
            time_t oldMTime = inode->i_mtime.tv_sec;
            loff_t oldSize = inode->i_size;

            __FhgfsOps_applyStatDataToInode(&kstat, inode);

            if( (inode->i_mtime.tv_sec != oldMTime) || (inode->i_size != oldSize) )
            { // todo: file contents changed => flush pages if we support mmap writes some
                  // time in the future (so just uncomment the following two lines)
               // Note: This would probably be wrong because we don't update the mtime during writes
               //    and coherence should be handled by the coherence protocol, so mtime/size
               //    shouldn't change without us knowing about it
               //if(!S_ISDIR(inode->i_mode) )
               //   invalidate_remote_inode(inode);
            }
         }
         else
         { // object type changed
            // Not good, inode has become a new object => limit damage by marking it as bad
            umode_t savedMode = inode->i_mode; // save mode
            make_bad_inode(inode);
            inode->i_mode = savedMode; // restore mode

            // todo: flush pages if we support mmap writes some time in the future
               // (so just uncomment the following two lines)
            if(!S_ISDIR(inode->i_mode) )
               invalidate_remote_inode(inode);

            retVal = -EREMOTEIO;
         }
      }

   } // end of (statRes == FhgfsOpsErr_SUCCESS) case


   return retVal;
}

/**
 * Note: This version is a wrapper for _refreshInodeUnlocked() that acquired the B-K-Lock.
 *
 * @param dentry the dentry for the inode that should be revalidated
 * @param path the resolved path (relative to mount point) for the dentry
 * @return 0 on success (validity), negative linux error code if no longer valid
 */
int __FhgfsOps_refreshInode(struct dentry* dentry, char* path)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   int retVal;

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   retVal = __FhgfsOps_refreshInodeUnlocked(dentry, path);

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK

   return retVal;
}


int __FhgfsOps_fillSuper(struct super_block* sb, void* rawMountOptions, int silent)
{
   //Logger* log = App_getLogger(Program_getApp() );
   //const char* logContext = "__FhgfsOps_fillSuper";

   struct inode* rootInode;
   struct dentry* rootDentry;
   struct kstat kstat;

   // init per-mount app object

   if(__FhgfsOps_initApp(sb, rawMountOptions) )
      return -EPERM;

   // set up super block data

   sb->s_maxbytes = MAX_LFS_FILESIZE;
   sb->s_blocksize = PAGE_CACHE_SIZE;
   sb->s_blocksize_bits = PAGE_CACHE_SHIFT;
   sb->s_magic = FHGFS_MAGIC;
   sb->s_op = &fhgfs_sops;
   sb->s_time_gran = 100;
   sb->s_flags |= MS_NODIRATIME;

   //Logger_logFormatted(log, 5, logContext, "MS_NOEXEC specified: %d",
   //   (int)(sb->s_flags & MS_NOEXEC) );


   // init root inode

   memset(&kstat, 0, sizeof(struct kstat) );

   kstat.ino = 2; /* traditional root inode number */
   kstat.mode = S_IFDIR | 0777; // allow access for everyone
   kstat.atime = kstat.mtime = kstat.ctime = current_fs_time(sb);
   kstat.uid = current->fsuid;
   kstat.gid = current->fsgid;
   kstat.blksize = PAGE_CACHE_SIZE;
   kstat.nlink = 1;

   rootInode = __FhgfsOps_newInode(sb, &kstat, 0);
   if(!rootInode)
   {
      __FhgfsOps_uninitApp(FhgfsOps_getApp(sb) );
      return -ENOMEM;
   }

   rootDentry = d_alloc_root(rootInode);
   if(!rootDentry)
   {
      __FhgfsOps_uninitApp(FhgfsOps_getApp(sb) );
      iput(rootInode);
      return -ENOMEM;
   }

   rootDentry->d_time = jiffies;
   sb->s_root = rootDentry;

   return 0;
}

void FhgfsOps_putSuper(struct super_block* sb)
{
   App* app = FhgfsOps_getApp(sb);

   if(app)
      __FhgfsOps_uninitApp(app);
}

void FhgfsOps_killSB(struct super_block* sb)
{
   // this is just a debug helper block (for dentry reference counters)
   /*
   printk_fhgfs(FHGFS_KERN_INFO, "calling autfs4_force_release\n"); // debug add
   __FhgfsOps_forceDentryRelease(sb); // debug add
   printk_fhgfs(FHGFS_KERN_INFO, "after autfs4_force_release\n"); // debug add
   */

   kill_anon_super(sb);
}

int FhgfsOps_registerFilesystem()
{
   return register_filesystem(&fhgfs_fs_type);
}

int FhgfsOps_unregisterFilesystem()
{
   return unregister_filesystem(&fhgfs_fs_type);
}

/**
 * Creates and initializes the per-mount application object.
 */
int __FhgfsOps_initApp(struct super_block* sb, char* rawMountOptions)
{
   MountConfig* mountConfig;
   fhgfs_bool parseRes;
   App* app;
   int appRes;

   // create mountConfig (parse from mount options)
   mountConfig = MountConfig_construct();

   parseRes = MountConfig_parseFromRawOptions(mountConfig, rawMountOptions);
   if(!parseRes)
   {
      MountConfig_destruct(mountConfig);
      return APPCODE_INVALID_CONFIG;
   }


   //printk_fhgfs(FHGFS_KERN_INFO, "Initializing App...\n"); // debug in

   app = App_construct(mountConfig);

   //printk_fhgfs(FHGFS_KERN_INFO, "Running App...\n"); // debug in

   appRes = App_run(app);

   if(appRes != APPCODE_NO_ERROR)
   { // error occured => clean up
      printk_fhgfs_debug(FHGFS_KERN_INFO, "Stopping App...\n");

      App_stop(app);

      printk_fhgfs_debug(FHGFS_KERN_INFO, "Cleaning up...\n");

      App_destruct(app);

      printk_fhgfs_debug(FHGFS_KERN_INFO, "App destructed.\n");

      return appRes;
   }

   FhgfsOps_setApp(sb, app);

   ProcFs_createEntries(app);

   // app running
   printk_fhgfs(FHGFS_KERN_INFO, "FhGFS mount ready.\n");

   return appRes;
}

/**
 * Stops and destroys the per-mount application object.
 */
void __FhgfsOps_uninitApp(App* app)
{
   //printk_fhgfs(FHGFS_KERN_INFO, "fhgfs_client preparing exit...\n");

   ProcFs_removeEntries(app);

   //printk_fhgfs(FHGFS_KERN_INFO, "Stopping App...\n"); // debug in

   App_stop(app);

   //printk_fhgfs(FHGFS_KERN_INFO, "Cleaning up...\n");

   App_destruct(app);

   //printk_fhgfs(FHGFS_KERN_INFO, "App destructed.\n");

   printk_fhgfs(FHGFS_KERN_INFO, "FhGFS unmounted.\n");
}

