#include <include_closed/toolkit/NoAllocBufferStore.h>
#include <include_closed/net/filesystem/FhgfsOpsRemoting.h>
#include "FhgfsOps_versions.h"
#include "FhgfsOps.h"


#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,16)

int FhgfsOps_statfs(struct super_block* sb, struct kstatfs* kstatfs)
{

#else

int FhgfsOps_statfs(struct dentry* dentry, struct kstatfs* kstatfs)
{
   struct super_block* sb = dentry->d_sb;

#endif // LINUX_VERSION_CODE


   App* app = FhgfsOps_getApp(sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_statfs";


   int retVal = -EREMOTEIO;
   FhgfsOpsErr statRes;

   int64_t sizeTotal = 0;
   int64_t sizeFree = 0;

   Logger_logFormatted(log, 5, logContext, "called.");

   memset(kstatfs, 0, sizeof(struct kstatfs) );

   __FhgfsOps_lockFhgfs(app); // B-K-LOCK

   kstatfs->f_type      = sb->s_magic;
   kstatfs->f_bsize     = 65536;
   kstatfs->f_namelen   = NAME_MAX;


   statRes = FhgfsOpsRemoting_statStoragePath(app, &sizeTotal, &sizeFree);

   LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %d", (int)statRes);

   retVal = FhgfsOpsErr_toSysErr(statRes);

   if(statRes == FhgfsOpsErr_SUCCESS)
   { // success => assign received values

      // note: do_div(n64, base32) assigns the result to n64 and returns the 32-bit remainder!
      kstatfs->f_blocks = sizeTotal;
      do_div(kstatfs->f_blocks, (int)kstatfs->f_bsize); // total blocks in the file system

      kstatfs->f_bfree = sizeFree;
      do_div(kstatfs->f_bfree, kstatfs->f_bsize); // free for superuser

      kstatfs->f_bavail    = kstatfs->f_bfree; // free for non-superuser

      retVal = 0;
   }

   __FhgfsOps_unlockFhgfs(app); // B-K-UNLOCK


   return retVal;
}



#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,17)

struct super_block* __FhgfsOps_getSB(struct file_system_type *fs_type,
   int flags, const char *dev_name, void *data)
{
   return get_sb_nodev(fs_type, flags, data, __FhgfsOps_fillSuper);
}

#else

int __FhgfsOps_getSB(struct file_system_type *fs_type,
   int flags, const char *dev_name, void *data, struct vfsmount *mnt)
{
   return get_sb_nodev(fs_type, flags, data, __FhgfsOps_fillSuper, mnt);
}

#endif // LINUX_VERSION_CODE


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)

/**
 * Releases dentry references (during unmount).
 *
 * Note: This is meant to be just a debug helper for dentry reference counters
 */
void __FhgfsOps_forceDentryRelease(struct super_block *sb)
{
   struct dentry* this_parent = sb->s_root;
   struct list_head* next;

   if (!sb->s_root)
      return;

   spin_lock(&dcache_lock);
repeat:
   next = this_parent->d_subdirs.next;
resume:
   while(next != &this_parent->d_subdirs)
   {
      struct dentry *dentry = list_entry(next, struct dentry, d_u.d_child);

      /* Negative dentry - don't care */
      if (!dentry->d_inode)
      {
         next = next->next;
         continue;
      }

      if(!list_empty(&dentry->d_subdirs) )
      {
         this_parent = dentry;
         goto repeat;
      }

      next = next->next;

      spin_unlock(&dcache_lock);

      printk_fhgfs(FHGFS_KERN_INFO, "cleaning up dentry %p %.*s (d_count: %lld)\n",
         dentry, (int)dentry->d_name.len, dentry->d_name.name,
         (long long)atomic_read(&dentry->d_count) ); // debug comment


      while(atomic_read(&dentry->d_count) >= 1) // we need >=1 here (not > 1)!
         dput(dentry);

      spin_lock(&dcache_lock);
   }

   if(this_parent != sb->s_root)
   {
      struct dentry *dentry = this_parent;

      next = this_parent->d_u.d_child.next;
      this_parent = this_parent->d_parent;

      spin_unlock(&dcache_lock);

      printk_fhgfs(FHGFS_KERN_INFO, "parent dentry %p %.*s (d_count: %lld)\n",
         dentry, (int)dentry->d_name.len, dentry->d_name.name,
         (long long)atomic_read(&dentry->d_count) ); // debug comment

      while(atomic_read(&dentry->d_count) > 1) // we need >1 here (not >= 1)!
         dput(dentry);

      spin_lock(&dcache_lock);

      goto resume;
   }

   spin_unlock(&dcache_lock);
}

#endif // LINUX_VERSION_CODE


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)

int FhgfsOps_follow_link(struct dentry* dentry, struct nameidata* nd)
{
   int retVal = 0;

#else

void* FhgfsOps_follow_link(struct dentry* dentry, struct nameidata* nd)
{
   void* retVal = NULL;

#endif // LINUX_VERSION_CODE

   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_follow_link";

   NoAllocBufferStore* bufStore = App_getPathBufStore(app);
   char* pathStoreBuf;
   char* path = __FhgfsOps_pathResolveToStoreBuf(bufStore, dentry, &pathStoreBuf);

   int readRes;
   char* bufPage = (char*)__os_get_free_page();
   char* destination = bufPage;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", path);

   readRes = FhgfsOpsHelper_readlink(app, path, bufPage, PAGE_SIZE-1);

   if(readRes < 0)
   {
      destination = ERR_PTR(readRes);
   }
   else
      bufPage[readRes] = 0;

   if(readRes == (PAGE_SIZE-1) )
   { // link destination too long
      destination = ERR_PTR(-ENAMETOOLONG);
   }

   // store link destination
   nd_set_link(nd, destination);

   // clean-up
   NoAllocBufferStore_addBuf(bufStore, pathStoreBuf);

   if(IS_ERR(destination) )
      os_free_page( (unsigned long)bufPage);

   // Note: free_page() is called by the put_link method in the success case

   return retVal;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)

void FhgfsOps_put_link(struct dentry *dentry, struct nameidata* nd)

#else

void FhgfsOps_put_link(struct dentry *dentry, struct nameidata* nd, void* p)

#endif // LINUX_VERSION_CODE

{
   char* destination = nd_get_link(nd);

   if(!IS_ERR(destination) )
      os_free_page( (unsigned long)destination);
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)

int FhgfsOps_flush(struct file *file)

#else

int FhgfsOps_flush(struct file *file, fl_owner_t id)

#endif // LINUX_VERSION_CODE


{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_flush";

   struct inode* inode = file->f_dentry->d_inode;
   int retVal = 0;

   Logger_logFormatted(log, 5, logContext, "called. Path: %s", "n/a");

   //printk_fhgfs_debug(FHGFS_KERN_INFO, "waiting for flush page writes\n"); // debug in (delete me)
   retVal = filemap_fdatawrite(inode->i_mapping);
   //printk_fhgfs_debug(FHGFS_KERN_INFO, "flush page writes complete\n"); // debug in (delete me)

   return retVal;
}

